import pyqrlew as pq
import numpy as np
from typing import List
import matplotlib.pylab as plt

DBNAME: str = 'pyqrlew-db'
USER: str = 'postgres'
PASSWORD: str = 'pyqrlew-db'
PORT: str = 5433
N_SIM = 500


class StochasticTester:
    def __init__(self, database, delta, privacy_unit, synthetic_data):
        self.database = database
        self.delta = delta
        self.privacy_unit = privacy_unit
        self.synthetic_data = synthetic_data

    def compute_dp_relation(self, query, epsilon):
        budget = {"epsilon": epsilon, "delta": self.delta}
        relation = self.database.dataset().sql(query)

        relation_with_privatequeries = relation.rewrite_with_differential_privacy(
            self.database.dataset(),
            self.synthetic_data,
            self.privacy_unit,
            budget
        )
        return relation_with_privatequeries.relation()

    def utility_per_epsilon(self, query: str, epsilons: List[float], n_sim: int):
        print('computing utility per epsilon ...')
        relations = {
            epsilon: self.compute_dp_relation(query, epsilon)
            for epsilon in epsilons
        }
        print('done')
        return {
            epsilon: [
                database.eval(dp_relation)[0][0]
                for _ in range(n_sim)
            ]
            for epsilon, dp_relation in relations.items()
        }

    def adjacent_databases(self, query: str, adj_query: str, epsilon: float, n_sim: int):
        print('test on adjacent databases ...')
        dp_relation = self.compute_dp_relation(query, epsilon)
        adj_dp_relation = self.compute_dp_relation(adj_query, epsilon)
        results = [
            database.eval(dp_relation)[0][0]
            for _ in range(n_sim)
        ]
        adj_results = [
            database.eval(adj_dp_relation)[0][0]
            for _ in range(n_sim)
        ]
        print('done')
        return results, adj_results


def plot_utility(dp_results: List[float], true_res: float):
    [
        plt.hist(results, bins=20, alpha=0.5, label = fr"$\varepsilon = {epsilon}$")
        for (epsilon, results) in dp_results.items()
    ]
    plt.axvline(true_res, color = 'red')
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.legend()


def plot_adjacent_histograms(array1: List[float], array2: List[float], epsilon:float, delta: float, bins:float=40, alpha=0.5, color1='blue', color2='orange'):
    xmin = min(min(array1), min(array2))
    xmax = max(max(array1), min(array2))
    bins = np.linspace(xmin, xmax, bins)
    values1, bins = np.histogram(array1, bins=bins)
    values1 = values1 / np.sum(values1)
    values2, _ = np.histogram(array2, bins=bins)
    values2 = values2 / np.sum(values2)
    x = bins[1:] + (bins[1] -bins[0])

    plt.bar(x, np.exp(epsilon) * values2 + delta, width=bins[1]-bins[0], alpha=alpha, color=color2, label=r'$e^{\varepsilon}P(f(\mathcal{D}_2)) + \delta$')
    plt.bar(x, values1, alpha=alpha, color=color1, width=bins[1]-bins[0], label=r'$P(f(\mathcal{D}_1))$')
    plt.grid(True, linestyle='--', alpha=0.7)
    plt.legend()


def run_test(stochastic_tester: StochasticTester, query: str, adj_query: str, name: str, n_sim: int):
    print(f"Query = {query}")
    dp_results = stochastic_tester.utility_per_epsilon(query, epsilons=[1., 5., 10.], n_sim=n_sim)
    true_res = database.execute(query)[0][0]
    plt.figure(figsize=(12, 5))
    plt.subplot(1, 2, 1)
    plot_utility(dp_results=dp_results, true_res=true_res)

    epsilon = 1.
    res0, res1 = stochastic_tester.adjacent_databases(query, adj_query, epsilon=epsilon, n_sim=n_sim)
    plt.subplot(1, 2, 2)
    plot_adjacent_histograms(res0, res1, epsilon=epsilon, delta=stochastic_tester.delta)

    plt.suptitle(query)
    plt.tight_layout()
    plt.savefig(name)
    print("================================")


def test_sum(stochastic_tester, n_sim):
    def write_inner_query(d, a, b):
        return f"SELECT {d} as x FROM testing.table1 WHERE {d} > {a} AND {d} < {b}"

    distributions = [
        ('laplace', -2.5, 2.5),
        ('gaussian', -2., 2.),
        ('uniform', -1., 1.),
        ('halton', -1., 1.),
    ]

    for (d, a, b) in distributions:
        inner_query = write_inner_query(d, a, b)
        query = f"SELECT SUM(x) AS my_sum FROM ({inner_query}) AS my_table"
        adj_query = f"SELECT SUM(x) AS my_sum FROM ({inner_query} AND id != 2) AS my_table"
        run_test(stochastic_tester, query=query, adj_query=adj_query, name=f'figures/sum_{d}.png', n_sim=n_sim)

def test_avg(stochastic_tester, n_sim):
    def write_inner_query(d, a, b):
        return f"SELECT {d} as x FROM testing.table1 WHERE {d} > {a} AND {d} < {b}"

    distributions = [
        ('laplace', -2.5, 2.5),
        ('gaussian', -2., 2.),
        ('uniform', -1., 1.),
        ('halton', -1., 1.),
    ]

    for (d, a, b) in distributions:
        inner_query = write_inner_query(d, a, b)
        query = f"SELECT AVG(x) AS my_sum FROM ({inner_query}) AS my_table"
        adj_query = f"SELECT AVG(x) AS my_sum FROM ({inner_query} AND id != 2) AS my_table"
        run_test(stochastic_tester, query=query, adj_query=adj_query, name=f'figures/avg_{d}.png', n_sim=n_sim)

def test_count(stochastic_tester, n_sim):
    query = f"SELECT COUNT(x) AS my_sum FROM (SELECT laplace AS x FROM testing.table1) AS my_table"
    adj_query = f"SELECT COUNT(x) AS my_sum FROM (SELECT laplace AS x FROM testing.table1 WHERE id != 2) AS my_table"
    run_test(stochastic_tester, query=query, adj_query=adj_query, name='figures/count.png', n_sim=n_sim)


if __name__ == "__main__":
    database = pq.stochatic_dataset.StochasticDatabase('testing', DBNAME, USER, PASSWORD, PORT)
    print('creating tables ...')
    ## create table
    column_specs = [
        pq.stochatic_dataset.ColumnSpec(name="laplace", distribution=pq.stochatic_dataset.Distribution.LAPLACE, loc=0, scale=1),
        pq.stochatic_dataset.ColumnSpec(name="uniform", distribution=pq.stochatic_dataset.Distribution.UNIFORM, low=-1, high=1),
        pq.stochatic_dataset.ColumnSpec(name="gaussian", distribution=pq.stochatic_dataset.Distribution.GAUSSIAN, loc=0, scale=1),
        pq.stochatic_dataset.ColumnSpec(name="halton", distribution=pq.stochatic_dataset.Distribution.HALTON, base=2, low=-1, high=1),
    ]
    database.create_table("table1", 100, column_specs, True)
    column_specs = [
        pq.stochatic_dataset.ColumnSpec(name="laplace", distribution=pq.stochatic_dataset.Distribution.LAPLACE, loc=0, scale=1),
        pq.stochatic_dataset.ColumnSpec(name="uniform", distribution=pq.stochatic_dataset.Distribution.UNIFORM, low=-1, high=1),
        pq.stochatic_dataset.ColumnSpec(name="gaussian", distribution=pq.stochatic_dataset.Distribution.GAUSSIAN, loc=0, scale=1),
        pq.stochatic_dataset.ColumnSpec(name="halton", distribution=pq.stochatic_dataset.Distribution.HALTON, base=2, low=-1, high=1),
    ]
    database.create_table("table2", 1000, column_specs, False)
    print('done')
    privacy_unit = [
        ("table1", [], "id"),
    ]
    synthetic_data = [
        (["testing", "table1"], ["testing", "table1"]),
    ]
    stochastic_tester = StochasticTester(database, 1 / 100 ** (3/2), privacy_unit, synthetic_data)
    test_sum(stochastic_tester, 1000)
    test_count(stochastic_tester, 1000)
    test_avg(stochastic_tester, 1000)










import logging
from uuid import uuid4 as generate_uuid
from typing import Optional, Tuple
import json
from sqlalchemy import MetaData, Table, Column, types, select, func
from sqlalchemy.engine import Engine
import pyqrlew as qrl

def dataset(name: str, engine: Engine, schema_name: Optional[str]=None) -> qrl.Dataset:
    metadata = MetaData()
    metadata.reflect(engine, schema=schema_name)

    def _dataset_schema_size() -> Tuple[dict, dict, Optional[dict]]:
        """Return a (dataset, schema) pair or (dataset, schema, size) triplet """
        ds = _dataset()
        return (
            ds,
            _schema(ds),
            _size(ds),
        )

    def _dataset() -> dict:
        """Return a (dataset, schema) pair or (dataset, schema, size) triplet """
        return {
            '@type': 'sarus_data_spec/sarus_data_spec.Dataset',
            'uuid': generate_uuid().hex,
            'name': 'Transformed',
            'spec': {
                'transformed': {
                    'transform': generate_uuid().hex,
                    'arguments': [],
                    'named_arguments': {},
                },
            },
            'properties': {},
            'doc': 'This ia a demo dataset for testing purpose',
        }

    def _schema(dataset: dict) -> dict:
        tables = {"fields": [_table(metadata.tables[name]) for name in metadata.tables]}

        if schema_name is not None:
            tables = {
                "fields": [{
                    'name': schema_name,
                    'type': {
                        'name': 'Union',
                        'union': tables,
                        'properties': {
                            'public_fields': '[]'
                        },
                    }
                }],
            }

        return {
            '@type': 'sarus_data_spec/sarus_data_spec.Schema',
            'uuid': generate_uuid().hex,
            'dataset': dataset['uuid'],
            'name': name,
            'type': {
                # Slugname
                'name': name.lower(),
                'struct': {
                    'fields': [
                        {
                            'name': 'sarus_data',
                            'type': {
                                'name': 'Union',
                                'union': tables,
                                'properties': {
                                    'public_fields': '[]',
                                },
                            },
                        },
                        {
                            'name': 'sarus_weights',
                            'type': {
                                'name': 'Integer',
                                'integer': {
                                    'min': '-9223372036854775808',
                                    'max': '9223372036854775807',
                                    'base': 'INT64',
                                    'possible_values': []
                                },
                                'properties': {}
                            },
                        },
                        {
                            'name': 'sarus_is_public',
                            'type': {
                                'name': 'Boolean',
                                'boolean': {},
                                'properties': {},
                            },
                        },
                        {
                            'name': 'sarus_protected_entity',
                            'type': {
                                'name': 'Id',
                                'id': {
                                    'base': 'STRING',
                                    'unique': False,
                                },
                                'properties': {},
                            },
                        },
                    ],
                },
                'properties': {}
            },
            'protected': {
                'label': 'data',
                'paths': [],
                'properties': {},
            },
            'properties': {
                'max_max_multiplicity': '1',
                'foreign_keys': '',
                'primary_keys': '',
            }
        }

    def _table(tab: Table) -> dict:
        return {
            'name': tab.name,
            'type': {
                'name': 'Struct',
                'struct': {
                    'fields': [_column(col) for col in tab.columns],
                },
                'properties': {},
            }
        }

    def _column(col: Column) -> dict:
        if isinstance(col.type, types.Integer) or isinstance(col.type, types.BigInteger):
            return {
                'name': col.name,
                'type': {
                    'name': 'Integer',
                    'integer': {
                        'base': 'INT64',
                        'min': '-9223372036854775808',
                        'max': '9223372036854775807',
                        'possible_values': [],
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Float) or isinstance(col.type, types.Numeric):
            return {
                'name': col.name,
                'type': {
                    'name': 'Float64',
                    'float': {
                        'base': 'FLOAT64',
                        'min': '-1.7976931348623157e+308',
                        'max': '1.7976931348623157e+308',
                        'possible_values': [],
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.String) or isinstance(col.type, types.Text) or isinstance(col.type, types.Unicode) or isinstance(col.type, types.UnicodeText):
            return {
                'name': col.name,
                'type': {
                    'name': 'Text UTF-8',
                    'text': {
                        'encoding': 'UTF-8'
                    },
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Boolean):
            return {
                'name': col.name,
                'type': {
                    'name': 'Boolean',
                    'boolean': {},
                    'properties': {},
                },
            }
        elif isinstance(col.type, types.Date) or isinstance(col.type, types.DateTime) or isinstance(col.type, types.Time):
            return {
                'name': col.name,
                'type': {
                    'name': 'Datetime',
                    'datetime': {
                        'format': '%Y-%m-%d %H:%M:%S',
                        'min': '01-01-01 00:00:00',
                        'max': '9999-12-31 00:00:00',
                    },
                    'properties': {},
                },
            }
        else:
            return {
                'name': col.name,
                'type': {
                    'name': 'Type',
                    'type': {},
                    'properties': {},
                },
            }

    def _size(dataset: dict) -> dict:
        tables = {'fields': [_table_size(metadata.tables[name]) for name in metadata.tables]}
        if schema_name is not None:
            tables = {
                'fields': [
                    {
                        'name': schema_name,
                        'statistics': {
                            'name': 'Union',
                            'union': tables,
                            'properties': {},
                        },
                        'properties': {},
                    },
                ],
            }

        return {
            '@type': 'sarus_data_spec/sarus_data_spec.Size',
            'uuid': generate_uuid().hex,
            'dataset': dataset['uuid'],
            'name': f'{name}_sizes',
            'statistics': {
                'name': 'Union',
                'union': tables,
                'properties': {},
            },
            'properties': {},
        }

    def _table_size(tab: Table) -> dict:
        with engine.connect() as conn:
            result = conn.execute(select(func.count()).select_from(tab))
            size = result.scalar()
        multiplicity = 1.0
        return {
            'name': tab.name,
            'statistics': {
                'name': 'Struct',
                'struct': {
                    'fields': [_column_size(col, size, multiplicity) for col in tab.columns],
                    'size': str(size),
                    'multiplicity': multiplicity,
                },
                'properties': {},
            }
        }

    def _column_size(col: Column, size: int, multiplicity: float) -> dict:
        if isinstance(col.type, types.Integer) or isinstance(col.type, types.BigInteger):
            return {
                'name': col.name,
                'statistics': {
                    "integer": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Integer",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.Float) or isinstance(col.type, types.Numeric):
            return {
                'name': col.name,
                'statistics': {
                    "float": {
                        "distribution": {
                            "double": {
                                "max": '1.7976931348623157e+308',
                                "min": '-1.7976931348623157e+308',
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Float",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.String) or isinstance(col.type, types.Text) or isinstance(col.type, types.Unicode) or isinstance(col.type, types.UnicodeText):
            return {
                'name': col.name,
                'statistics': {
                    "text": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Text",
                    "properties": {}
                }
            }

        elif isinstance(col.type, types.Boolean):
            return {
                'name': col.name,
                'statistics': {
                    "boolean": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Boolean",
                    "properties": {}
                }
            }
        elif isinstance(col.type, types.Date) or isinstance(col.type, types.DateTime) or isinstance(col.type, types.Time):
            return {
                'name': col.name,
                'statistics': {
                    "datetime": {
                        "distribution": {
                            "integer": {
                                "max": "9223372036854775807",
                                "min": "-9223372036854775808",
                                "points": []
                            },
                            "properties": {}
                            },
                        "multiplicity": multiplicity,
                        "size": size
                    },
                    "name": "Datetime",
                    "properties": {}
                }
            }

        else: # TODO: support more types
            raise NotImplementedError(f"SQL -> Sarus Conversion not supported for {col.type} SQL type")
    # Gather protobufs
    dataset, schema, size = _dataset_schema_size()
    # Display when debugging
    logging.debug(json.dumps(dataset))
    logging.debug(json.dumps(schema))
    logging.debug(json.dumps(size))
    # Return the result
    return qrl.Dataset(json.dumps(dataset), json.dumps(schema), json.dumps(size))